/*
 * include/asm-xtensa/namei.h
 *
 * Included from linux/fs/namei.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 - 2005 Tensilica Inc.
 */

#ifndef _XTENSA_NAMEI_H
#define _XTENSA_NAMEI_H

#ifdef __KERNEL__

/* This dummy routine maybe changed to something useful
 * for /usr/gnemul/ emulation stuff.
 * Look at asm-sparc/namei.h for details.
 */

#define __emul_prefix() NULL

#endif	/* __KERNEL__ */
#endif	/* _XTENSA_NAMEI_H */
